#include <nv/common.hh>
#include <iomanip>
#include <nv/gfx/keyframed_mesh.hh>
#include <nv/interface/vertex_buffer.hh>
#include <nv/gl/gl_device.hh>
#include <nv/gfx/image.hh>
#include <nv/interface/context.hh>
#include <nv/interface/window.hh>
#include <nv/interface/program.hh>
#include <nv/interface/texture2d.hh>
#include <nv/interface/mesh_loader.hh>
#include <nv/io/c_file_system.hh>
#include <nv/formats/md3_loader.hh>
#include <nv/profiler.hh>
#include <nv/logging.hh>
#include <nv/logger.hh>
#include <nv/math.hh>
#include <nv/time.hh>
#include <nv/string.hh>
#include <nv/interface/mesh.hh>
#include <glm/gtx/rotate_vector.hpp>
#include <glm/gtc/matrix_access.hpp>
#include <glm/gtx/matrix_interpolation.hpp>

bool GPU_ANIMATION = false;

class mesh_part
{
public:
	mesh_part( const std::string& path, nv::program* program, nv::window* window )
	{
		NV_PROFILE("mesh_construct");
		nv::md3_loader* loader;
		{
			NV_PROFILE("loader->load");
			nv::c_file_system fs;
			nv::stream* mesh_file = fs.open( path.c_str() );
			loader = new nv::md3_loader();
			loader->load( *mesh_file );
			delete mesh_file;
		}

		{
			NV_PROFILE("create_mesh_data");
			m_data      = loader->release_mesh_data();
		}
		delete loader;

		{
			NV_PROFILE("create_mesh");
			if ( GPU_ANIMATION )
				m_mesh      = new nv::keyframed_mesh_gpu( window->get_context(), m_data, program );
			else
				m_mesh      = new nv::keyframed_mesh_cpu( window->get_context(), m_data );
		}

	}

	nv::mat4 get_transform( const std::string& tag )
	{
		return m_mesh->get_tag( tag ).extract();
	}

	void setup_animation( nv::uint32 start, nv::uint32 stop, nv::uint32 fps, bool loop )
	{
		m_mesh->setup_animation( start, stop, fps, loop );
	}

	void update( nv::uint32 ms )
	{
		m_mesh->update( ms );
	}

	nv::mesh_interface* get_mesh() const { return m_mesh; }

	size_t get_max_frames()
	{
		return m_mesh->get_max_frames();
	}

	void draw( nv::context* c, nv::program* program, nv::render_state& rstate, const glm::mat4& m, const glm::mat4& v, const glm::mat4& p )
	{
		NV_PROFILE( "mesh-draw" );
		glm::mat4 mv = v * m;
		program->set_opt_uniform( "nv_m_modelview", mv );
		program->set_opt_uniform( "nv_m_normal", glm::transpose(glm::inverse(glm::mat3(mv))) );
		program->set_uniform( "matrix_mvp", p * mv );
		
		c->draw( rstate, program, m_mesh );
	}

	~mesh_part()
	{
		delete m_data;
		delete m_mesh;
	}

private:
	nv::mesh_data*           m_data;
	nv::keyframed_mesh*      m_mesh;
};

class application
{
public:
	application();
	bool initialize();
	bool run();
	~application();
protected:
	nv::device*       m_device;
	nv::window*       m_window;
	nv::texture2d*    m_diffuse;
	nv::texture2d*    m_diffuse_weapon;
	nv::clear_state   m_clear_state;
	nv::render_state  m_render_state;

	mesh_part* m_torso;
	mesh_part* m_legs;
	mesh_part* m_head;
	mesh_part* m_weapon;
	nv::program*      m_program;
};

application::application()
{
	NV_PROFILE( "app_construct" );
	m_device = new nv::gl_device();
	m_window = m_device->create_window( 800, 600, false );

	nv::sampler sampler( nv::sampler::LINEAR, nv::sampler::REPEAT );
	nv::image_data* data = m_device->create_image_data( "data/doom.png" );
	m_diffuse  = m_device->create_texture2d( data->get_size(), nv::RGBA, nv::UBYTE, sampler, (void*)data->get_data() );
	delete data;
	data = m_device->create_image_data( "data/rocketl.png" );
	m_diffuse_weapon = m_device->create_texture2d( data->get_size(), nv::RGBA, nv::UBYTE, sampler, (void*)data->get_data() );
	delete data;

	m_clear_state.buffers = nv::clear_state::COLOR_AND_DEPTH_BUFFER;
	m_render_state.depth_test.enabled = true;
	m_render_state.culling.enabled    = true;
	m_render_state.blending.enabled   = false;
	m_render_state.blending.src_rgb_factor   = nv::blending::SRC_ALPHA;
	m_render_state.blending.dst_rgb_factor   = nv::blending::ONE_MINUS_SRC_ALPHA;
	m_render_state.blending.src_alpha_factor = nv::blending::SRC_ALPHA;
	m_render_state.blending.dst_alpha_factor = nv::blending::ONE_MINUS_SRC_ALPHA;
}

bool application::initialize()
{
	NV_PROFILE( "app_initialize" );
	m_program = m_device->create_program( 
		nv::slurp( GPU_ANIMATION ? "obj_gpu.vert" : "obj_cpu.vert" ), 
		nv::slurp( "obj.frag" ) 
	);
	m_torso   = new mesh_part( "data/upper.md3", m_program, m_window );
	m_legs    = new mesh_part( "data/lower.md3", m_program, m_window );
	m_head    = new mesh_part( "data/head.md3", m_program, m_window );
	m_weapon  = new mesh_part( "data/rocketl.md3", m_program, m_window );
	return true;
}

bool application::run()
{
	nv::profiler::pointer()->log_report(); 
	NV_PROFILE( "app_run" );
	int keypress = 0;

	glm::vec3 move( 0, 0, 0 );

	nv::uint32 ticks   = m_device->get_ticks();
	nv::uint32 last_ticks;
	nv::fps_counter_class< nv::system_us_timer > fps_counter;

	m_legs->setup_animation( 0, m_legs->get_max_frames(), 20, true );
	m_torso->setup_animation( 0, m_torso->get_max_frames(), 20, true );

	bool common = true;
	while(!keypress) 
	{
		last_ticks = ticks;
		ticks      = m_device->get_ticks();
		nv::uint32 elapsed = ticks - last_ticks;
		m_torso->update( elapsed );
		m_legs->update( elapsed );
		{
			NV_PROFILE( "clear" );
			m_window->get_context()->clear( m_clear_state );
		}

		glm::mat4 view;
		glm::mat4 projection;
		{
			NV_PROFILE( "update_sh" );

			glm::vec3 source( 80.0f, 0.0f, 0.0f );
			glm::vec3 eye = glm::rotate( source, (ticks / 20.f), glm::vec3( 0.0,1.0,0.0 ) );

			view       = glm::lookAt(eye + move, glm::vec3(0.0f, 0.0f, 0.0f) + move, glm::vec3(0.0, 1.0, 0.0));
			projection = glm::perspective(60.0f, 1.0f*800.0f/600.0f, 0.1f, 1000.0f);

			m_diffuse->bind( 0 );
			m_program->set_opt_uniform( "nv_m_projection", projection );
			m_program->set_uniform( "light_position", glm::vec3(120.0, 120.0, 0) );
			m_program->set_uniform( "light_diffuse",  glm::vec4(1.0,1.0,1.0,1.0) );
			m_program->set_uniform( "light_specular", glm::vec4(1.0,1.0,1.0,1.0) );
			m_program->set_uniform( "diffuse", 0 );
		}

		{
			NV_PROFILE( "draw" );
			glm::mat4 model      = glm::mat4(1.0f);
			m_legs->draw( m_window->get_context(), m_program, m_render_state, model, view, projection );

			//model = m_legs->get_transform( "tag_torso", last_legs_frame, legs_frame, legs_interpolate );
			model = m_legs->get_transform( "tag_torso" );
			m_torso->draw( m_window->get_context(), m_program, m_render_state, model, view, projection );

			glm::mat4 head = model * m_torso->get_transform( "tag_head" ); //, last_torso_frame, torso_frame, torso_interpolate );
			m_head->draw( m_window->get_context(), m_program, m_render_state, head, view, projection );

			glm::mat4 weapon = model * m_torso->get_transform( "tag_weapon" ); //, last_torso_frame, torso_frame, torso_interpolate );
			m_diffuse_weapon->bind( 0 );
			m_weapon->draw( m_window->get_context(), m_program, m_render_state, weapon, view, projection );

		}

		{
			NV_PROFILE( "swap" );
			m_window->swap_buffers();
		}

		{
			NV_PROFILE( "pollevent" );
			nv::io_event event;
			while(m_window->poll_event(event)) 
			{      
				switch (event.type) 
				{
				case nv::EV_QUIT:
					keypress = 1;
					break;
				case nv::EV_KEY:
					if (event.key.pressed)
					{
						switch (event.key.code) 
						{
						case nv::KEY_ESCAPE : keypress = 1; break;
						case nv::KEY_1 : m_legs->setup_animation( 0, 89, 20, true ); m_torso->setup_animation( 0, 89, 20, true ); common = true; break;
						case nv::KEY_2 : m_legs->setup_animation( 90, 8, 20, true ); if (common) m_torso->setup_animation( 151, 1, 20, true ); common = false; break;
						case nv::KEY_3 : m_legs->setup_animation( 98, 12, 20, true ); if (common) m_torso->setup_animation( 151, 1, 20, true ); common = false; break;
						case nv::KEY_4 : m_legs->setup_animation( 110, 9, 15, true ); if (common) m_torso->setup_animation( 151, 1, 20, true ); common = false; break;
						case nv::KEY_5 : m_legs->setup_animation( 138, 10, 15, true ); if (common) m_torso->setup_animation( 151, 1, 20, true ); common = false; break;
						case nv::KEY_6 : m_legs->setup_animation( 171, 9, 15, true ); if (common) m_torso->setup_animation( 151, 1, 20, true ); common = false; break;
						case nv::KEY_Q : m_torso->setup_animation( 90, 40, 20, true ); if (common) m_legs->setup_animation( 171, 10, 15, true ); common = false; break;
						case nv::KEY_W : m_torso->setup_animation( 130, 6, 15, true ); if (common) m_legs->setup_animation( 171, 10, 15, true ); common = false; break;
						case nv::KEY_E : m_torso->setup_animation( 136, 6, 15, true ); if (common) m_legs->setup_animation( 171, 10, 15, true ); common = false; break;
						case nv::KEY_R : m_torso->setup_animation( 142, 5, 20, true ); if (common) m_legs->setup_animation( 171, 10, 15, true ); common = false; break;
						case nv::KEY_T : m_torso->setup_animation( 147, 4, 20, true ); if (common) m_legs->setup_animation( 171, 10, 15, true ); common = false; break;
						case nv::KEY_Y : m_torso->setup_animation( 151, 1, 15, true ); if (common) m_legs->setup_animation( 171, 10, 15, true ); common = false; break;
						case nv::KEY_U : m_torso->setup_animation( 152, 1, 15, true ); if (common) m_legs->setup_animation( 171, 10, 15, true ); common = false; break;
						case nv::KEY_F1 : nv::profiler::pointer()->log_report(); break;
						default: break;
						}
					}
					break;
				default: break;
				}
			}
		}
		fps_counter.tick();
	}
	return true;
}

application::~application()
{
	delete m_program;
	delete m_torso;
	delete m_legs;
	delete m_head;
	delete m_weapon;
	delete m_diffuse;
	delete m_diffuse_weapon;
	delete m_window;
	delete m_device;
}


int main(int, char* [])
{
	nv::logger log(nv::LOG_TRACE);
	log.add_sink( new nv::log_file_sink("log.txt"), nv::LOG_TRACE );
	log.add_sink( new nv::log_console_sink(), nv::LOG_TRACE );
	
	NV_LOG( nv::LOG_NOTICE, "Logging started" );
	application app;
	if ( app.initialize() )
	{
		app.run();
	}
	NV_LOG( nv::LOG_NOTICE, "Logging stopped" );

	return 0;
}

